@@ -27,6 +27,10 @@
  * See the ChangeLog files for a list of changes. 
  */
 
+#ifdef HAVE_CONFIG_H
+#include <config.h>
+#endif
+
 #include <string.h>
 
 #include <gconf/gconf-client.h>
@@ -1609,6 +1613,7 @@ gedit_prefs_manager_get_encodings (void)
 
 	 	res = g_slist_reverse (res);
 	}
+
 	return res;
 }
 
