@@ -28,13 +28,17 @@
  * See the ChangeLog files for a list of changes. 
  */
 
+#ifdef HAVE_CONFIG_H
+#include <config.h>
+#endif
+
 #include <glib.h>
 #include <stdlib.h>
 #include <string.h>
 
 #include "gedit-undo-manager.h"
 #include "gedit-debug.h"
-#include "gedit-prefs.h"
+#include "gedit-prefs-manager.h"
 #include "gedit-marshal.h"
 
 typedef struct _GeditUndoAction  		GeditUndoAction;
@@ -43,7 +47,7 @@ typedef struct _GeditUndoDeleteAction		GeditUndoDeleteAction;
 
 typedef enum {
 	GEDIT_UNDO_ACTION_INSERT,
-	GEDIT_UNDO_ACTION_DELETE,
+	GEDIT_UNDO_ACTION_DELETE
 } GeditUndoActionType;
 
 /* 
@@ -56,6 +60,7 @@ struct _GeditUndoInsertAction
 	gint   pos; 
 	gchar *text;
 	gint   length;
+	gint   chars;
 };
 
 struct _GeditUndoDeleteAction
@@ -92,6 +97,8 @@ struct _GeditUndoManagerPrivate
 	gboolean	can_redo;
 
 	gint		running_not_undoable_actions;
+
+	gint		num_of_groups;
 };
 
 enum {
@@ -121,7 +128,6 @@ static void gedit_undo_manager_check_list_size 		(GeditUndoManager *um);
 
 static gboolean gedit_undo_manager_merge_action 	(GeditUndoManager *um, 
 		                                         GeditUndoAction *undo_action);
-static gint gedit_undo_manager_get_number_of_groups 	(GeditUndoManager *um);
 
 static GObjectClass 	*parent_class 				= NULL;
 static guint 		undo_manager_signals [LAST_SIGNAL] 	= { 0 };
@@ -205,6 +211,8 @@ gedit_undo_manager_init (GeditUndoManager *um)
 	um->priv->can_redo = FALSE;
 
 	um->priv->running_not_undoable_actions = 0;
+
+	um->priv->num_of_groups = 0;
 }
 
 static void
@@ -278,7 +286,8 @@ gedit_undo_manager_new (GeditDocument* document)
 	return um;
 }
 
-void gedit_undo_manager_begin_not_undoable_action (GeditUndoManager *um)
+void 
+gedit_undo_manager_begin_not_undoable_action (GeditUndoManager *um)
 {
 	gedit_debug (DEBUG_UNDO, "");
 
@@ -288,19 +297,50 @@ void gedit_undo_manager_begin_not_undoable_action (GeditUndoManager *um)
 	++um->priv->running_not_undoable_actions;
 }
 
-void gedit_undo_manager_end_not_undoable_action (GeditUndoManager *um)
+static void 
+gedit_undo_manager_end_not_undoable_action_internal (GeditUndoManager *um)
 {
 	gedit_debug (DEBUG_UNDO, "");
 
 	g_return_if_fail (GEDIT_IS_UNDO_MANAGER (um));
 	g_return_if_fail (um->priv != NULL);
 
-	if (um->priv->running_not_undoable_actions <= 0)
-		um->priv->running_not_undoable_actions = 0;
-	else
-		--um->priv->running_not_undoable_actions;
+	g_return_if_fail (um->priv->running_not_undoable_actions > 0);
+	
+	--um->priv->running_not_undoable_actions;
 }
 
+void 
+gedit_undo_manager_end_not_undoable_action (GeditUndoManager *um)
+{
+	gedit_debug (DEBUG_UNDO, "");
+
+	g_return_if_fail (GEDIT_IS_UNDO_MANAGER (um));
+	g_return_if_fail (um->priv != NULL);
+
+	gedit_undo_manager_end_not_undoable_action_internal (um);
+
+	if (um->priv->running_not_undoable_actions == 0)
+	{	
+		gedit_undo_manager_free_action_list (um);
+	
+		um->priv->next_redo = -1;	
+
+		if (um->priv->can_undo)
+		{
+			um->priv->can_undo = FALSE;
+			g_signal_emit (G_OBJECT (um), undo_manager_signals [CAN_UNDO], 0, FALSE);
+		}
+
+		if (um->priv->can_redo)
+		{
+			um->priv->can_redo = FALSE;
+			g_signal_emit (G_OBJECT (um), undo_manager_signals [CAN_REDO], 0, FALSE);
+		}
+	}
+}
+
+
 gboolean
 gedit_undo_manager_can_undo (const GeditUndoManager *um)
 {
@@ -346,18 +386,27 @@ gedit_undo_manager_undo (GeditUndoManager *um)
 		switch (undo_action->action_type)
 		{
 			case GEDIT_UNDO_ACTION_DELETE:
+				gedit_document_set_cursor (
+					um->priv->document, 
+					undo_action->action.delete.start);
+
 				gedit_document_insert_text (
 					um->priv->document, 
 					undo_action->action.delete.start, 
 					undo_action->action.delete.text,
 					strlen (undo_action->action.delete.text));
-				break;
+
+								break;
 			case GEDIT_UNDO_ACTION_INSERT:
 				gedit_document_delete_text (
 					um->priv->document, 
 					undo_action->action.insert.pos, 
 					undo_action->action.insert.pos + 
-						undo_action->action.insert.length); 
+						undo_action->action.insert.chars); 
+
+				gedit_document_set_cursor (
+					um->priv->document, 
+					undo_action->action.insert.pos);
 				break;
 
 			default:
@@ -367,7 +416,7 @@ gedit_undo_manager_undo (GeditUndoManager *um)
 
 	} while (undo_action->order_in_group > 1);
 
-	gedit_undo_manager_end_not_undoable_action (um);
+	gedit_undo_manager_end_not_undoable_action_internal (um);
 	
 	if (!um->priv->can_redo)
 	{
@@ -407,9 +456,18 @@ gedit_undo_manager_redo (GeditUndoManager *um)
 					um->priv->document, 
 					undo_action->action.delete.start, 
 					undo_action->action.delete.end); 
+
+				gedit_document_set_cursor (
+					um->priv->document,
+					undo_action->action.delete.start);
+
 				break;
 				
 			case GEDIT_UNDO_ACTION_INSERT:
+				gedit_document_set_cursor (
+					um->priv->document,
+					undo_action->action.insert.pos);
+
 				gedit_document_insert_text (
 					um->priv->document, 
 					undo_action->action.insert.pos, 
@@ -432,7 +490,7 @@ gedit_undo_manager_redo (GeditUndoManager *um)
 		
 	} while ((undo_action != NULL) && (undo_action->order_in_group > 1));
 
-	gedit_undo_manager_end_not_undoable_action (um);
+	gedit_undo_manager_end_not_undoable_action_internal (um);
 
 	if (um->priv->next_redo < 0)
 	{
@@ -468,9 +526,20 @@ gedit_undo_manager_free_action_list (GeditUndoManager *um)
 	{
 		GeditUndoAction *undo_action = 
 			(GeditUndoAction *)(g_list_nth_data (um->priv->actions, n));
-	
+
+		gedit_debug (DEBUG_UNDO, "Free action (type %s) %d/%d", 
+				(undo_action->action_type == GEDIT_UNDO_ACTION_INSERT) ? "insert":
+				"delete", n, len);
+
 		if (undo_action->action_type == GEDIT_UNDO_ACTION_INSERT)
-			g_free (undo_action->action.insert.text);	
+			g_free (undo_action->action.insert.text);
+		else if (undo_action->action_type == GEDIT_UNDO_ACTION_DELETE)
+			g_free (undo_action->action.delete.text);
+		else
+			g_return_if_fail (FALSE);
+
+		if (undo_action->order_in_group == 1)
+			--um->priv->num_of_groups;
 
 		g_free (undo_action);
 	}
@@ -490,17 +559,16 @@ gedit_undo_manager_insert_text_handler (GtkTextBuffer *buffer, GtkTextIter *pos,
 	if (um->priv->running_not_undoable_actions > 0)
 		return;
 
+	g_return_if_fail (strlen (text) == (guint)length);
+	
 	undo_action.action_type = GEDIT_UNDO_ACTION_INSERT;
 
 	undo_action.action.insert.pos    = gtk_text_iter_get_offset (pos);
-	undo_action.action.insert.text   = g_strdup (text);
+	undo_action.action.insert.text   = (gchar*) text;
 	undo_action.action.insert.length = length;
-	
-	/* FIXME: it is broken with non ASCII chars */
-	/*
-	if ((g_utf8_strlen (text, length) > 1) || (g_utf8_get_char (text) == '\n'))
-	*/
-	if ((length > 1) || (g_utf8_get_char (text) == '\n'))
+	undo_action.action.insert.chars  = g_utf8_strlen (text, length);
+
+	if ((undo_action.action.insert.chars > 1) || (g_utf8_get_char (text) == '\n'))
 
 	       	undo_action.mergeable = FALSE;
 	else
@@ -522,6 +590,8 @@ gedit_undo_manager_delete_range_handler (GtkTextBuffer *buffer, GtkTextIter *sta
 
 	undo_action.action_type = GEDIT_UNDO_ACTION_DELETE;
 
+	gtk_text_iter_order (start, end);
+
 	undo_action.action.delete.start  = gtk_text_iter_get_offset (start);
 	undo_action.action.delete.end    = gtk_text_iter_get_offset (end);
 
@@ -530,17 +600,21 @@ gedit_undo_manager_delete_range_handler (GtkTextBuffer *buffer, GtkTextIter *sta
 						undo_action.action.delete.start,
 						undo_action.action.delete.end);
 
-	/* FIXME: it is broken with non ASCII chars */
-	/*
-	if ((undo_action.action.delete.end - undo_action.action.delete.start) > 1)
-	*/
-	if ((strlen (undo_action.action.delete.text) > 1) ||
-	    (g_utf8_get_char (undo_action.action.delete.text  ) == '\n'))
+	if (((undo_action.action.delete.end - undo_action.action.delete.start) > 1) ||
+	     (g_utf8_get_char (undo_action.action.delete.text  ) == '\n'))
 	       	undo_action.mergeable = FALSE;
 	else
 		undo_action.mergeable = TRUE;
 
+	gedit_debug (DEBUG_UNDO, "START: %d", undo_action.action.delete.start);
+	gedit_debug (DEBUG_UNDO, "END: %d", undo_action.action.delete.end);
+	gedit_debug (DEBUG_UNDO, "TEXT: %s", undo_action.action.delete.text);
+
 	gedit_undo_manager_add_action (um, undo_action);
+
+	g_free (undo_action.action.delete.text);
+
+	gedit_debug (DEBUG_UNDO, "DONE");
 }
 
 static void 
@@ -587,9 +661,22 @@ gedit_undo_manager_add_action (GeditUndoManager *um, GeditUndoAction undo_action
 	{
 		action = g_new (GeditUndoAction, 1);
 		*action = undo_action;
-	
+
+		if (action->action_type == GEDIT_UNDO_ACTION_INSERT)
+			action->action.insert.text = g_strdup (undo_action.action.insert.text);
+		else if (action->action_type == GEDIT_UNDO_ACTION_DELETE)
+			action->action.delete.text = g_strdup (undo_action.action.delete.text); 
+		else
+		{
+			g_free (action);
+			g_return_if_fail (FALSE);
+		}
+		
 		++um->priv->actions_in_current_group;
 		action->order_in_group = um->priv->actions_in_current_group;
+
+		if (action->order_in_group == 1)
+			++um->priv->num_of_groups;
 	
 		um->priv->actions = g_list_prepend (um->priv->actions, action);
 	}
@@ -608,6 +695,7 @@ gedit_undo_manager_add_action (GeditUndoManager *um, GeditUndoAction undo_action
 		g_signal_emit (G_OBJECT (um), undo_manager_signals [CAN_REDO], 0, FALSE);
 	}
 
+	gedit_debug (DEBUG_UNDO, "DONE");
 }
 
 static void 
@@ -635,6 +723,9 @@ gedit_undo_manager_free_first_n_actions (GeditUndoManager *um, gint n)
 		else
 			g_return_if_fail (FALSE);
 
+		if (undo_action->order_in_group == 1)
+			--um->priv->num_of_groups;
+
 		g_free (undo_action);
 
 		um->priv->actions = g_list_delete_link (um->priv->actions, um->priv->actions);
@@ -647,15 +738,19 @@ gedit_undo_manager_free_first_n_actions (GeditUndoManager *um, gint n)
 static void 
 gedit_undo_manager_check_list_size (GeditUndoManager *um)
 {
+	gint undo_levels;
+	
 	gedit_debug (DEBUG_UNDO, "TODO");
 
 	g_return_if_fail (GEDIT_IS_UNDO_MANAGER (um));
 	g_return_if_fail (um->priv != NULL);
 	
-	if (gedit_settings->undo_levels < 1)
+	undo_levels = gedit_prefs_manager_get_undo_actions_limit ();
+	
+	if (undo_levels < 1)
 		return;
 
-	if (gedit_undo_manager_get_number_of_groups (um) > gedit_settings->undo_levels)
+	if (um->priv->num_of_groups > undo_levels)
 	{
 		GeditUndoAction *undo_action;
 		GList* last;
@@ -672,6 +767,9 @@ gedit_undo_manager_check_list_size (GeditUndoManager *um)
 			else
 				g_return_if_fail (FALSE);
 
+			if (undo_action->order_in_group == 1)
+				--um->priv->num_of_groups;
+
 			g_free (undo_action);
 
 			um->priv->actions = g_list_delete_link (um->priv->actions, last);
@@ -681,8 +779,7 @@ gedit_undo_manager_check_list_size (GeditUndoManager *um)
 			undo_action = (GeditUndoAction*) last->data;
 
 		} while ((undo_action->order_in_group > 1) || 
-			 (gedit_undo_manager_get_number_of_groups (um) > 
-			  gedit_settings->undo_levels));
+			 (um->priv->num_of_groups > undo_levels));
 	}	
 }
 
@@ -700,8 +797,6 @@ gedit_undo_manager_check_list_size (GeditUndoManager *um)
 static gboolean 
 gedit_undo_manager_merge_action (GeditUndoManager *um, GeditUndoAction *undo_action)
 {
-	/* FIXME: it is broken with non ASCII chars */
-	
 	GeditUndoAction *last_action;
 	
 	gedit_debug (DEBUG_UNDO, "");
@@ -738,12 +833,15 @@ gedit_undo_manager_merge_action (GeditUndoManager *um, GeditUndoAction *undo_act
 			gchar *str;
 			
 #define L  (last_action->action.delete.end - last_action->action.delete.start - 1)
+#define g_utf8_get_char_at(p,i) g_utf8_get_char(g_utf8_offset_to_pointer((p),(i)))
+		
+			gedit_debug (DEBUG_UNDO, "L = %d", L);
 
 			/* Deleted with the delete key */
-			if ((undo_action->action.delete.text [0] != ' ') &&
-			    (undo_action->action.delete.text [0] != '\t') &&
-                            ((last_action->action.delete.text [L] == ' ') ||
-			     (last_action->action.delete.text [L] == '\t')))
+			if ((g_utf8_get_char (undo_action->action.delete.text) != ' ') &&
+			    (g_utf8_get_char (undo_action->action.delete.text) != '\t') &&
+                            ((g_utf8_get_char_at (last_action->action.delete.text, L) == ' ') ||
+			     (g_utf8_get_char_at (last_action->action.delete.text, L)  == '\t')))
 			{
 				last_action->mergeable = FALSE;
 				return FALSE;
@@ -753,7 +851,7 @@ gedit_undo_manager_merge_action (GeditUndoManager *um, GeditUndoAction *undo_act
 				undo_action->action.delete.text);
 			
 			g_free (last_action->action.delete.text);
-			last_action->action.delete.end += 1;
+			last_action->action.delete.end += (undo_action->action.delete.end - undo_action->action.delete.start);
 			last_action->action.delete.text = str;
 		}
 		else
@@ -761,10 +859,10 @@ gedit_undo_manager_merge_action (GeditUndoManager *um, GeditUndoAction *undo_act
 			gchar *str;
 			
 			/* Deleted with the backspace key */
-			if ((undo_action->action.delete.text [0] != ' ') &&
-			    (undo_action->action.delete.text [0] != '\t') &&
-                            ((last_action->action.delete.text [0] == ' ') ||
-			     (last_action->action.delete.text [0] == '\t')))
+			if ((g_utf8_get_char (undo_action->action.delete.text) != ' ') &&
+			    (g_utf8_get_char (undo_action->action.delete.text) != '\t') &&
+                            ((g_utf8_get_char (last_action->action.delete.text) == ' ') ||
+			     (g_utf8_get_char (last_action->action.delete.text) == '\t')))
 			{
 				last_action->mergeable = FALSE;
 				return FALSE;
@@ -781,13 +879,15 @@ gedit_undo_manager_merge_action (GeditUndoManager *um, GeditUndoAction *undo_act
 	else if (undo_action->action_type == GEDIT_UNDO_ACTION_INSERT)
 	{
 		gchar* str;
-
+		
+#define I (last_action->action.insert.chars - 1)
+		
 		if ((undo_action->action.insert.pos != 
-		     	(last_action->action.insert.pos + last_action->action.insert.length)) ||
-		    ((undo_action->action.insert.text [0] != ' ') &&
-		      (undo_action->action.insert.text [0] != '\t') &&
-		     ((last_action->action.insert.text [last_action->action.insert.length - 1] == ' ') ||
-		      (last_action->action.insert.text [last_action->action.insert.length - 1] == '\t')))
+		     	(last_action->action.insert.pos + last_action->action.insert.chars)) ||
+		    ((g_utf8_get_char (undo_action->action.insert.text) != ' ') &&
+		      (g_utf8_get_char (undo_action->action.insert.text) != '\t') &&
+		     ((g_utf8_get_char_at (last_action->action.insert.text, I) == ' ') ||
+		      (g_utf8_get_char_at (last_action->action.insert.text, I) == '\t')))
 		   )
 		{
 			last_action->mergeable = FALSE;
@@ -800,6 +900,8 @@ gedit_undo_manager_merge_action (GeditUndoManager *um, GeditUndoAction *undo_act
 		g_free (last_action->action.insert.text);
 		last_action->action.insert.length += undo_action->action.insert.length;
 		last_action->action.insert.text = str;
+		last_action->action.insert.chars += undo_action->action.insert.chars;
+
 	}
 	else
 		g_warning ("Unknown action inside undo merge encountered");	
@@ -807,21 +909,4 @@ gedit_undo_manager_merge_action (GeditUndoManager *um, GeditUndoAction *undo_act
 	return TRUE;
 }
 
-static gint
-gedit_undo_manager_get_number_of_groups (GeditUndoManager *um)
-{
-	GList *list = um->priv->actions;
-  	gint num = 0;
-
-	gedit_debug (DEBUG_UNDO, "");
-  
-  	while (list)
-    	{
-      		if (((GeditUndoAction* ) list->data)->order_in_group == 1)
-			++num;
-      		list = list->next;
-    	}
-  
-  	return num;
-}
 
